<?php

/*    This file is part of Zzzt!.
    Copyright (C) 2009-2019  Antonio Ho, José Jorge

    Zzzt! is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Zzzt! is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Zzzt! .  If not, see <http://www.gnu.org/licenses/>.
*/

// error_reporting(E_ALL);
require '../i18n/switch.php';
require '../inverter/class_db.php';

class inverter
{
  public $history=array();//table with history of freq updates
  public $status=array();//table with latest values obtained from the inverter
  public $events=array();
  public $config,$parameters,$frequent_updates;
  public $efficiency; //efficiency
  public $connected=false; //is the inverter online
   
  function __construct($model=''){
    require '../config/config.php';
    $this->config=$config;
    if($model!='') $this->config['model']=$model;
    require "../model/{$this->config['model']}.php";//we load the model data
    $this->events=$events;
    $this->parameters=$parameters;
    $this->frequent_updates=$frequent_updates;
    $this->crontab_monitor=$crontab_monitor;
    $this->last_day=$last_day;

    //tables are empty if the inverter is offline
    foreach($parameters as $k=>$v)
            $this->status[$k]=0;
    foreach($frequent_updates as $k=>$v)
            $this->status[$k]=0;
    foreach($events as $k=>$v)
            $this->status[$k]=$v;
    $this->status['timestamp']=time();
  }
  
  public function connected(){
    if( $this->connected == true ) return ONLINE;
    else return OFFLINE;
  }
  
  // write to config/local_config.php to current config
  public function fwrite_config(){
    $myFile = "../config/local_config.php";
    $fh = fopen($myFile, 'w') or die("can't open file");
    $stringData = "<?php
      \$config=array();\n";
    foreach ( $this->config as $k => $v )
      $stringData.="\$config['$k']='$v';\n";
    $stringData .= "?>";
    fwrite($fh, $stringData);
    fclose($fh);
  }
  
  // updates the crontab table if the inverter is online
  // intended to standalone_crontab
  public function crontab(){
    $online=$this->update($this->crontab_monitor);
  }
  
  // updates the frequent table if the inverter is online
  // and all other tables if more than 1 minute has passed
  public function refresh(){
    $online=$this->update($this->frequent_updates);
    $this->status['EFFICIENCY']=sprintf('%.2f',$this->status['PAC']/($this->status['PDC']+0.0001)*100);
    if($online) {
      $this->history[]=$this->status;
      if(($this->status['THR']!=date('H')) ||
          ($this->status['TMI']+5) == date('i')){ //only updates every 5 minutes
        $this->update($this->parameters);
        $this->updateEvents();
      }
    }
    //we limit the history to last 60 frequent_updates
    if( count($this->history) >=60 ) array_shift($this->history);
  }

  // returns an html table with the status of $what_table
  public function table_status($what_table) {
          $echo='<table border=1 width="100%"><tr>';
          foreach ($what_table as $nom => $parameter){
                  if($nom=='SYS') $valeur=constant($this->status[$nom]);
                  else $valeur = $this->status[$nom];
                  $unit = $what_table[$nom]["unit"];
                  $titre = constant($nom);
                  $echo.='<tr><td>'.$titre.'</td>
                                  <td>'.$valeur.$unit.'</td>
                          </tr>';
          }
          return $echo.'</tr></table>';
  }

  // returns an html table with the $events last events
  public function table_events($events) {
          $counter=0;
          $echo='<table border=1 width="100%"><tr>';
          foreach ($this->events as $n => $p){
                  $counter++;
                  $echo.='<tr>
                          <td>'.date('H:i:s(D d)',$this->status[$n]['timestamp']).'</td>
                          <td>'.constant($this->status[$n]['event']).'</td>
                          </tr>';
          }
          return $echo.'</tr></table>';
  }

  // builds the query in the language of the inverter from a table of queries
  private function compose_sm($message){
    $checksum=0;

    $question="|".dechex($this->config['destnumber']).":".$message;
    $question=substr($question,0,-1)."|";
    $l=substr("0".dechex(14+strlen($question)),-2);
    $question=strtoupper(substr("0".dechex($this->config['sendnumber']),-2).";".substr("0".dechex($this->config['inverternumber']),-2).";".$l.$question);
    for($i=0; $i < strlen($question); $i++) {$checksum += ord($question[$i]);}
    $checksum=strtoupper(substr("000".dechex($checksum),-4));
    $question="{".$question.$checksum."}";
    return $question;
  }
  
  //Extracts the data from the inverter answer and write it in the status table
  // Returns : OK or an ERROR MESSAGE
  private function extract_sm($reponse) {
    if(($reponse[0]!="{") || ($reponse[strlen($reponse)-1]!="}"))
      return "Syntax Error {}";
    if (strlen($reponse) != hexdec(substr($reponse,7,2)))
      return "Lenght Error";
    $reponse = substr($reponse,1,-1);

    // Verify Checksum
    $checksum = hexdec(substr($reponse,-4));
    $reponse = substr($reponse,0,-4);
    for($i=0; $i < strlen($reponse); $i++) {$checksum -= ord($reponse[$i]);}
    if ($checksum != 0) return "Checksum Error";
  
    // Verify that it's the good inverter that has answered
    if (hexdec(substr($reponse,0,2)) != $this->config['inverternumber']) return "Bad Inverter Error";
    if (hexdec(substr($reponse,3,2)) != $this->config['sendnumber'] ) return "Destination Error";
  
    $reponse = substr($reponse,8);
    if (($reponse[0] !="|") || (substr($reponse,-1)!= "|"))
      return "Syntax of || error";
    $reponse=substr($reponse,1,-1);
  
    // Verify that it is DATA, not Messages
    $i = strpos($reponse,":");
    if ($i===FALSE) return "Syntax Error of Data Type";
  
    // Verify DATA ports
    $aut_ports =array("64","78","C8"); // autorised ports
    if (substr($reponse,0,$i)=="3E8") return "Inverter Error ".$reponse;
    if (!in_array(substr($reponse,0,$i),$aut_ports)) return "Data Port Error ".substr($reponse,0,$i-1);
    $reponse= substr($reponse,$i+1);

    // Here we have only the DATA, we fill an array with it
    $status = explode(";", $reponse);
    return $status;
  }

  private function connect() {
    $error = NULL;
    $attempts = 0;
    $timeout = 100 * $this->config['timeout'];  // adjust because we are sleeping in 10 milliseconds increments
    $connOK = 0;
    
    $socket = socket_create(AF_INET, SOCK_STREAM, SOL_TCP);
    if ($socket === false)
	throw new Exception(socket_strerror(socket_last_error())); 

    socket_set_nonblock($socket);
    $error = NULL;
    while (!($connOK = @socket_connect($socket, $this->config['ipaddress'], $this->config['serviceport'])) && $attempts++ < $timeout) {
	$error = socket_last_error($socket);
	if ($error == SOCKET_EISCONN)
	{
	    $connOK = true;
	    break;
	}

	if ($error != SOCKET_EINPROGRESS && $error != SOCKET_EALREADY) {
	    socket_close($socket);
	    throw new Exception(socket_strerror($error));
	}
	usleep(10000);
    }
    if (!$connOK) {
	$error = "Error Connecting Socket: Connect Timed Out After ".$this->config['timeout']." seconds: ".socket_strerror(socket_last_error($socket));
	socket_close($socket);
	throw new Exception($error);
    }
      
    socket_set_block($socket);
	    
    return $socket;
  }


  // Updates the inverter status, if the connection is possible
  private function update($what_table) {
          try {
                  $socket=$this->connect();
                  $message='';
                  foreach($what_table as $d => $parameter)
                  {
                          $message.=$d.";";
                  }
                  $q=$this->compose_sm($message);
                  socket_write($socket, $q, strlen($q));
                  $o = socket_read($socket, 2048);
                  socket_close($socket);
                  $status=$this->extract_sm($o);
                  if ( is_array($status) ) {
                        foreach ($status as $val) $this->updateStatus($what_table,$val);
                        $this->status['timestamp']=time();
                        // the inverter is back online
                        if ($this->connected==false) {
                                $this->connected=true;
                        }
                  }

          } catch (Exception $e) {
                  $this->connected=false;
          }
          return $this->connected;
  }
  
  
  
  // converts the hexadecimal date to SQL friendly
  private function mktimeFromSputnik($sdate,$stime)
  {
    $in_hour=hexdec($stime);
    return mktime(
      substr('0'.floor($in_hour/3600),-2),
      substr('0'.floor(fmod($in_hour,3600)/60),-2),
      substr('0'.fmod($in_hour,60),-2),
      sprintf("%02s",hexdec(substr($sdate,-4,2))),
      sprintf("%02s",hexdec(substr($sdate,-2))),
      hexdec(substr($sdate,0,-4))
      );
  }
  //updates the events 4 by 4 to prevent communication problems observed
  // with some inverter firmwares
  private function updateEvents() {
    $limit=0;//the inverter crash if event queries are too big
    foreach($this->events as $k=>$v) {
      $limit++;
      $subEvent[$k]=$v;
      if($limit>4) {$limit=0;$this->update($subEvent);$subEvent=array();}
    }
    if(sizeof($subEvent)>0) $this->update($subEvent);
  }

  //update differently the status table according to some exceptions
  private function updateStatus($what_table,$val) {
    list($name,$value)=explode("=",$val);
    if($name=='SYS') $value=substr($value,0,4);//delete the ,0
    if ($name=="DD01")
    {
       $dd=explode(",", $value);
       $value=$dd[1];
    }
    
    if(substr($name,0,2)=='EC') {
      list($sdate,$stime,$event)=explode(",",$value);
      $this->status[$name]=array(
                          'timestamp'=>$this->mktimeFromSputnik($sdate,$stime),
                          'event'=>hexdec($event)
                  );
    } else {
      $this->status[$name] = (string)(hexdec($value))*$what_table[$name]["coef"];}
  }
}
?>
