<?php
/*    This file is part of Zzzt!.
    Copyright (C) 2009-2019  Antonio Ho, José Jorge

    Zzzt! is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Zzzt! is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Zzzt! .  If not, see <http://www.gnu.org/licenses/>.

  ----------------------------------------------------
              Solarmax inverter simulator
  ----------------------------------------------------
  
  - needs php-cli because it is launched with php simu2.php
  - was made to be able to test Zzzt! at night ;-)
  - also understands the EXIT command that stops the simulator (as Ctrl+c)
*/

error_reporting(E_ALL ^ E_WARNING);
/* Runs forever */
set_time_limit(0);
ob_implicit_flush();

define ('DEF_NOM','Solarmax SIM');
define('DEF_ADDR','192.168.0.10');
define('DEF_PORT',12345);
define('DEF_INT',2);

$address = DEF_ADDR;
$int_addr = DEF_INT;
$port = DEF_PORT;

function extract_sm($reponse, $ad_onduleur, $ad_dest,&$valeurs)
{
  if(($reponse[0]!="{") || ($reponse[strlen($reponse)-1]!="}")) return "Erreur de syntaxe {}";
  if (strlen($reponse) != hexdec(substr($reponse,7,2))) return "Erreur de longueur";
  $reponse = substr($reponse,1,-1);
  $checksum = hexdec(substr($reponse,-4));
  $reponse = substr($reponse,0,-4);
  for($i=0; $i < strlen($reponse); $i++) {$checksum -= ord($reponse[$i]);}
  if ($checksum != 0) return "Erreur de checksum";
  if (hexdec(substr($reponse,0,2)) != $ad_onduleur) return "Erreur pas le bon onduleur";
  if (hexdec(substr($reponse,3,2)) != $ad_dest) return "Erreur de destinataire";
  $reponse = substr($reponse,8);
  if (($reponse[0] !="|") || (substr($reponse,-1)!= "|")) return "erreur de syntaxe ||";
  $reponse=substr($reponse,1,-1);
  $i = strpos($reponse,":");
  if ($i===FALSE) return "Erreur de syntaxe type de données";
  $aut_ports =array("64","78","C8"); // ports autorisés
  if (substr($reponse,0,$i)=="3E8") return "Erreur onduleur ".$reponse;
  if (!in_array(substr($reponse,0,$i),$aut_ports)) return "Erreur port de données ".substr($reponse,0,$i-1);
  $reponse= substr($reponse,$i+1);
  $valeurs = explode(";", $reponse);
return "OK";
}

function compose_sm($ad_onduleur, $ad_dest, $port, $data)
{
  $question="|";
  if (count($data)) $question.= dechex($port);
  $question.=':';
  foreach($data as $d)
  {
    $question.=$d.";";
  }
  $question=substr($question,0,-1)."|";
  $l=substr("0".dechex(14+strlen($question)),-2);
  $question=strtoupper(substr("0".dechex($ad_dest),-2).";".substr("0".dechex($ad_onduleur),-2).";".$l.$question);
  $checksum=0;
  for($i=0; $i < strlen($question); $i++) {$checksum += ord($question[$i]);}
  $checksum=strtoupper(substr("000".dechex($checksum),-4));
  $question="{".$question.$checksum."}";
  return $question;
}

function setval($row) {
  return dechex(rand($row->min*$row->coef, $row->max*$row->coef));
}

function setdate(){
  $dd=explode('-',date('Y-m-d'));
  $date=dechex($dd[0]).substr('0'.dechex($dd[1]),-2).substr('0'.dechex($dd[2]),-2);
  $hh=explode(':',date('H:i:s'));
  $heure=dechex($hh[0]*3600+$hh[1]*60+$hh[2]);
  return $date.",".$heure;
}

function setsys()
{
  $stat=array();
// all known real world codes
  for($i=1;$i<14;$i++) $stat[]='2'.sprintf('%04s',$i);
  for($i=110;$i<113;$i++) $stat[]='2'.sprintf('%04s',$i);
  for($i=114;$i<120;$i++) $stat[]='2'.sprintf('%04s',$i);
  for($i=122;$i<128;$i++) $stat[]='2'.sprintf('%04s',$i);
  for($i=130;$i<133;$i++) $stat[]='2'.sprintf('%04s',$i);
  for($i=134;$i<135;$i++) $stat[]='2'.sprintf('%04s',$i);

  return dechex($stat[rand(0,count($stat)-1)]).',0';
}

function setevent()
{
return dechex(date('Y')).
      sprintf("%02s",dechex(date('m'))).
      sprintf("%02s",dechex(date('d'))).','.
      sprintf("%04s",dechex(rand(21600,79200))).','.setsys();//between 6AM and 22PM in seconds since 00h00m00s
}

function ond_par($cde)
{
global $int_addr;
$equiv= array("ADR"=>$int_addr,"TYP"=>20020);

return dechex($equiv[$cde]);
}

function validateIpAddress($ip_addr)
{
  //first of all the format of the ip address is matched
  if(preg_match("/^(\d{1,3})\.(\d{1,3})\.(\d{1,3})\.(\d{1,3})$/",$ip_addr))
  {
     $parts=explode(".",$ip_addr);
    //now we need to check each part can range from 0-255
    foreach($parts as $ip_parts)
    {
      if(intval($ip_parts)>255 || intval($ip_parts)<0)
      return false;
    }
    return true;
  }
  else
    return false;
}

function print_aide()
{
echo <<<FIN


Usage : php Simu2.php [Options]

[Options] :
        --help, -h, -help, -? for help
        -ip 192.168.1.1 : IP adress of the simulator
        -po 12345 : TCP port of the simulator (1-65535)
        -ai 2 : internal address of the simulator (1-249)

Ex : php simu2.php -ip 192.168.1.10 -po 12345 -ai 2


FIN;
}

if (realpath($_SERVER['argv'][0]) == __FILE__ ) array_shift($argv);
$argcount =count($argv);
if ($argcount==1)
{
	print_aide();
	exit;
}

for ($i=0;$i<$argcount-1;$i+=2)
{
	switch ($argv[$i])
	{
		case '-ip':
			$address=$argv[$i+1];
			if(!validateIpAddress($address))
			{
				print_aide();
				exit;
			}
			break;
		case '-po' :
			$port=$argv[$i+1];
			if($port<1 || $port > 65535)
			{
				print_aide();
				exit;
			}
			break;
		case '-ai' :
			$int_addr=$argv[$i+1];
			if ($int_addr<1 || $int_addr > 249)
			{
				print_aide();
				exit;
			}
			break;
		default :
			print_aide();
			exit;
	}
}

 if (!extension_loaded('pdo_sqlite')) exit("ERROR : php pdo_sqlite extension is missing\n"); 

try
{
	$dbHandle = new PDO('sqlite:db/simubd.sqlite3');
}
catch( PDOException $exception )
{
	die($exception->getMessage());
}

$prepCde=$dbHandle->prepare("SELECT * FROM simu WHERE nom = ?");

echo "\n-----------------------------------------------------------------------\n
             SOLARMAX SIM (inverter simulator)\n
-----------------------------------------------------------------------\n";
echo ("\nInverter : ".DEF_NOM." with internal address : ".$int_addr." on IP : ".$address.", port : ".$port."\n");


$buf=NULL;
if (($sock = socket_create(AF_INET, SOCK_STREAM, SOL_TCP)) === false) {
    echo "socket_create() failed : " . socket_strerror(socket_last_error()) . "\n";
}

if (socket_bind($sock, $address, $port) === false) {
    echo "socket_bind() failed : " . socket_strerror(socket_last_error($sock)) . "\n";
}

if (socket_listen($sock, 5) === false) {
    echo "socket_listen() failed : " . socket_strerror(socket_last_error($sock)) . "\n";
}

do {
    if (($msgsock = socket_accept($sock)) === false) {
        echo "socket_accept() failed : " . socket_strerror(socket_last_error($sock)) . "\n";
        break;
    }

    do {
        // waits for a '{'        
        $buffer=NULL;
        if (false === ($buf = socket_read($msgsock, 1,PHP_NORMAL_READ))) {
            break;
        }
        if ($buf!=='{'){ continue ;}
        $buffer=$buf;
        // reads everything till the end
        do {
          if (false === ($buf = socket_read($msgsock, 1,PHP_BINARY_READ)))
          {
            break 2;
          }
          $buffer.=$buf;
        } while ($buf!="}");

// message analysis

        $cde=array();
        $result=extract_sm($buffer,251,$int_addr,$cde);
        if (!($result=='OK')){
          echo $result."\n";
          socket_write($msgsock, $result, strlen($result));
          continue;
        }
        if ($cde[0] == 'EXIT') {
            $msg="Exit command, server stopped\n";
            socket_write($msgsock, $msg, strlen($msg));
            socket_close($msgsock);
            break 2;
        }

// pour chaque commande, prépare une réponse
 
        $retour=NULL;
        for ($ii=0;$ii<count($cde);$ii++)
		{
		$prepCde->execute(array($cde[$ii]));

// chercher la commande parmi celles connues

          if (!$ligne2=$prepCde->fetch(PDO::FETCH_OBJ)){
            echo "Unknown command : -".$cde[$ii]."- please update the database.\n";
          }
          else {
            echo "Asked for {$cde[$ii]} \n";

// préparer la réponse
             switch($ligne2->type){
              case 'D':
                $retour[$ii]= $ligne2->nom."=".setdate();
                break;
              case 'E':
                $retour[$ii]=$ligne2->nom."=".setevent();
                break;
              case 'S':
                $retour[$ii]=$ligne2->nom."=".setsys();
                break;
              case 'ON';
                $retour[$ii]=$ligne2->nom."=".ond_par($ligne2->nom);
                break;
              default:
                $retour[$ii]= $ligne2->nom."=".setval($ligne2);
		if($ligne2->nom == 'PAC') {
			$pac=setval($ligne2);
                	$retour[$ii]= $ligne2->nom."=".$pac;
		}
		if($ligne2->nom == 'PDC') {
			$pdc=dechex(hexdec($pac)*(rand(100,190)*0.01));
			$retour[$ii]= $ligne2->nom."=".$pdc;
		}
            }
          }     
        }
          $msg=compose_sm(251,$int_addr,100,$retour);
          socket_write($msgsock,$msg,strlen($msg));
        //}
    } while (true);
    socket_close($msgsock);
} while (true);

socket_close($sock);
echo "---------------------------- SERVER DOWN ---------------------------------\n";
?>
